package za.co.eskom.nrs.xmlvend.client;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import nu.xom.Builder;
import nu.xom.Document;
import org.apache.log4j.Logger;
import za.co.eskom.nrs.xmlvend.client.comms.BasicSOAP11WSITransportLayer;
import za.co.eskom.nrs.xmlvend.client.comms.TransportLayer;
import za.co.eskom.nrs.xmlvend.client.setup.ClientSetupListener;
import za.co.eskom.nrs.xmlvend.client.wsdl.UseCaseMap;
import za.co.eskom.nrs.xmlvend.client.wsdl.UseCaseMap.UseCase;
import za.co.solms.framework.config.Configurable;
import za.co.solms.framework.config.Configuration;
import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.soap.SOAPMessage;

/** Used for use-case invocation (primary controller).
 * The Servlet Path "Info" indicates the use case name,
 * and for GET this is used to send a view back to the client, and for
 * POST is used to send the message as SOAP to the endpoint server,
 * receive a response, and hand that response to a pre-written
 * response view.
 * 
 * @author dawidl
 * @version 0.1 (Dec 19, 2005)
 *
 */
public class UseCaseServlet extends HttpServlet implements Configurable
{

	@Override
	public void init() throws ServletException
	{
		// Setup configuration
		setConfiguration( (Configuration)getServletContext().getAttribute(ClientSetupListener.CONFIGURATION) );
		
		// Get a handle to the use case map
		this.useCases = (UseCaseMap)
		getServletContext().getAttribute( ClientSetupListener.SERVICE_USE_CASES );
		
		// Check if it exists
		if (useCases == null)
		{
			throw new ServletException("No use-case map in context ("+ 
					ClientSetupListener.SERVICE_USE_CASES +")");
		}
		
		// Setup transport layer
		this.transport = new BasicSOAP11WSITransportLayer( getConfiguration() );
	}
	
	
	@Override
	protected void doGet(HttpServletRequest request, HttpServletResponse response) 
	throws ServletException, IOException
	{
		// Get use case
		UseCase u = getRequestedUseCase( request );
		
		// Get the input view (if one exists) and dispatch client to there
		String viewURI = u.getInputView();
		if (viewURI == null)
		{
			throw new ServletException(new UseCaseNotSupportedException( u.getName() ));
		}
		logger.debug("Presenting view to user: " + viewURI);
		
		request.getRequestDispatcher(viewURI).forward(request, response);
	}
	
	
	@Override
	protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
	{
		// Get use case
		UseCase u = getRequestedUseCase( request );
		
		// Parse POST data as XML
		logger.debug("Parsing XML Message data");
		Document d = null;
		try
		{
			d = xomBuilder.build(request.getInputStream());
		} 
		catch ( Exception e)
		{
			// If parsing fails, tell client
			response.sendError(HttpServletResponse.SC_BAD_REQUEST );
			return;
		}
		
		// TODO Validate message?
		//...
		
		// Store message information of last message for possible advice messages later
		saveAsLastSentMessage( d, request );
		
		// If configuration is per se, perform loopback to client (and end processing)
		if (Boolean.parseBoolean( getConfiguration().getProperty("xmlvend.debug.messageLoopback.loopback") ))
		{
			logger.info("Message loopback: Sending message back to client");
			sendXMLToClient( d, response );
			return;
		}
		
		// Send to endpoint server
		SOAPMessage sres = null;
		try
		{
			sres = this.transport.invokeService( new XMLMessage( d ) );
		}
		catch (IOException e)
		{
			throw new ServletException( new CommunicationFault(e) );
		}
		catch (Exception g)
		{
			throw new ServletException(g);
		}
		
		// If we have no response, throw an exception
		if (sres == null || sres.getMessage() == null)
		{
			logger.debug("No response returned");
			throw new ServletException( new NoServerResponseFault() );
		}
		
		// If response is a fault, throw an exception
		if (sres.isFault())
		{
			logger.debug("Response message is a fault");
			throw new ServletException( new SOAPFault( sres ) );
		}
		
		
		// Send to generic UI (for transformation and display)
		request.setAttribute("responseMessage", sres.getMessage().getRootElement().toXML());
		request.getRequestDispatcher("/useCaseViews/_anyResponse-out.jspx").forward(request, response);
	}
	
	
	/** Locate the use case for the given request */
	private UseCase getRequestedUseCase( HttpServletRequest request ) throws ServletException
	{
		// Get use case name (ignore first char, which is assumed to be "/")
		String ucName = request.getPathInfo();
		if (ucName.startsWith("/"))
		{
			ucName = ucName.substring(1);
		}
		logger.debug("Use case requested: " + ucName);
		
		
		// Look up by name, if it doesn't exist, throw an error
		UseCase u = useCases.getUseCaseByName( ucName );
		if (u == null)
		{
			throw new ServletException(new UseCaseNotSupportedException(ucName));
		}
		
		return u;
	}
	
	
	/** Stores the given message information in the current session. Currently,
	 * this is stored as a String until we can sort out the DOM vs XOM debate */
	protected void saveAsLastSentMessage( Document msg, HttpServletRequest request )
	{
		request.getSession().setAttribute( LAST_SENT_REQUEST, msg.toXML() );
		logger.debug("Stored message (string) as ${sessionScope."+ LAST_SENT_REQUEST +"}");
	}
	
	
	/** Sends the given XML to the client. No forther data should be sent after this (buffer is flushed) */
	protected void sendXMLToClient(Document xml, HttpServletResponse response) throws IOException
	{
		byte[] xmlData = xml.toXML().getBytes();
		response.setContentLength( xmlData.length );
		response.setContentType("text/xml");
		response.getOutputStream().write( xmlData );
		response.flushBuffer();
	}
	
	
	public void setConfiguration(Configuration config)
	{
		this.config = config;
	}
	
	
	public Configuration getConfiguration()
	{
		return config;
	}

	
	private UseCaseMap useCases;
	private Logger logger = Logger.getLogger(UseCaseServlet.class);
	private Builder xomBuilder = new Builder();
	private TransportLayer transport;
	private Configuration config;
	/** Key used to store the last sent request message in the HTTP request */
	public static final String LAST_SENT_REQUEST = "lastSentRequest";
}
