package za.co.eskom.nrs.xmlvend.client.comms;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.methods.ByteArrayRequestEntity;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.log4j.Logger;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.ParsingException;
import za.co.solms.framework.config.Configurable;
import za.co.solms.framework.config.Configuration;
import za.co.solms.framework.ooxml.XMLMessage;
import za.co.solms.framework.ooxml.soap.SOAPMessage;
import za.co.solms.framework.ooxml.soap.SOAPMessageFactory;

/** A basic transport layer, which makes use of a shared Configuration
 * to invoke a remote web service.
 * 
 * 'xmlvend.transport.service.url' - URL of ultimate receiver
 * 'xmlvend.transport.compression.request' - (boolean) compress requests
 * 'xmlvend.transport.compression.response' - (boolean) support compressed responses
 * 
 * Uses the Jakarta Commons HTTPClient for HTTP communication.
 * 
 * @author dawidl
 * @version 0.1 (Dec 20, 2005)
 *
 */
public class BasicSOAP11WSITransportLayer implements TransportLayer, Configurable
{
	public BasicSOAP11WSITransportLayer( Configuration config ) throws IllegalArgumentException
	{
		if (config == null)
			throw new IllegalArgumentException("Invalid configuration: null");
		setConfiguration( config );
	}

	public SOAPMessage invokeService(XMLMessage request) throws HttpException, IOException, Exception
	{
		String url = getServerURL();
		logger.debug("Request to send message to: " + url);
		
		// Wrap request in SOAP if it's not already SOAP
		logger.debug("Wrapping message in SOAP");
		SOAPMessage soapReq = (request instanceof SOAPMessage) ? (SOAPMessage)request : 
			SOAPMessageFactory.getInstance().createMessage( request );
		
		// Create POST request
		String xml = soapReq.getSOAPEnvelope().toXML();
		byte[] requestData = xml.getBytes();
		logger.debug("Message length: " + xml.length());
		PostMethod post = new PostMethod( url );
		
		// Soap Action (must be empty
		post.setRequestHeader("SOAPAction","");
		
		// Request compression
		if (shouldCompressRequests())
		{
			logger.debug("Enabling request message compression");
			logger.debug("Original message size: " + requestData.length + " bytes");
			// Create new set of (compressed) request data
			ByteArrayOutputStream bout = new ByteArrayOutputStream( requestData.length );
			GZIPOutputStream zout = new GZIPOutputStream( bout , 1024);
			zout.write( requestData );
			zout.flush();
			zout.finish();
			zout.close();
			// Replace original data with compressed (reference)
			requestData = bout.toByteArray();
			post.setRequestHeader("Content-Encoding","gzip");
			logger.debug("Compressed message size: " + requestData.length + " bytes");
		}
		
		// Set request content
		post.setRequestEntity(new ByteArrayRequestEntity( requestData, "text/xml") );
		
		// Response Compression
		if (supportsCompressedResponses())
		{
			logger.debug("Indicating that compressed responses are supported");
			post.setRequestHeader("Accept-Encoding","gzip");
		}
		
		// Invoke and disconnect
		HttpClient client = new HttpClient();
		Document responseDocument = null;
		// TODO: Setup HTTPClient properly, including logging, etc
		//...
		try
		{
			logger.debug("Sending message");
			client.executeMethod( post );
			InputStream fromServer = post.getResponseBodyAsStream();
			// If we support compressed responses, check if the response is compressed,
			// and wrap in a suitable stream
			Header ce = post.getResponseHeader("Content-Encoding");
			if (ce != null)
			{
				// Check if it contains GZIP
				String encoding = ce.getValue().toLowerCase();
				if (encoding.contains("gzip"))
				{
					// Wrap input stream with an inflator
					if (supportsCompressedResponses())
					{
						logger.debug("Server sent gzip-encoded response, setting up decompression");
						fromServer = new GZIPInputStream( fromServer, 32000 );
					}
					else
					{
						// We don't support it by configuration
						logger.error("Server sent compressed response, even though this client indicated that it does not support it");
						throw new IOException("Server sent compressed response (but support is disabled)");
					}
				}
			}
			//debug response message
			logger.info("HTTP Response Body = "  + fromServer.toString());
			// Parse XML
			responseDocument = xomBuilder.build( fromServer );
		} 
		catch (ParsingException pe)
		{
			logger.error("Server sent invalid XML", pe);
			throw pe;
		}
		catch (HttpException he)
		{
			//...
			logger.error("Message transmission failed", he);
			throw he;
		}
		catch (IOException ie)
		{
			//...
			logger.error("Message transmission failed", ie);
			throw ie;
		}
		finally
		{
			logger.debug("Releasing connection to remote server");
			post.releaseConnection();
		}
		
		SOAPMessage response = new SOAPMessage( responseDocument );
		return response;
	}
	
	public void setConfiguration(Configuration conf)
	{
		this.config = conf;
	}
	
	public Configuration getConfiguration()
	{
		return config;
	}
	
	private String getServerURL() throws Exception
	{
		String url = this.config.getProperty("xmlvend.transport.service.url");
		if (url == null || url.length() == 0)
			throw new Exception("No URL configured for endpoint server (xmlvend.transport.service.url)");
		return url;
	}
	
	private boolean supportsCompressedResponses()
	{
		String b = this.config.getProperty("xmlvend.transport.compression.response");
		return Boolean.parseBoolean(b);
	}
	
	private boolean shouldCompressRequests()
	{
		String b = this.config.getProperty("xmlvend.transport.compression.request");
		return Boolean.parseBoolean(b);
	}
	
	private Configuration config;
	private Builder xomBuilder = new Builder();
	private static Logger logger = Logger.getLogger(BasicSOAP11WSITransportLayer.class);
}
