package za.co.eskom.nrs.xmlvend.client.setup;

import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import za.co.eskom.nrs.xmlvend.client.MessageIDGenerator;
import za.co.eskom.nrs.xmlvend.client.wsdl.UseCaseMap;
import za.co.solms.framework.config.Configuration;
import za.co.solms.framework.config.FileConfiguration;

/** Sets up the client (context, use cases, views, etc). Binds a "UseCaseMap"
 * to the servlet context as 'service.useCases', and sets up configuration
 * and various utilities. */

public class ClientSetupListener implements ServletContextListener
{
	public void contextInitialized(ServletContextEvent evt)
	{
		// Initialise Logging
		BasicConfigurator.configure();
		
		// Load use cases
		loadUseCases( evt.getServletContext() );
		
		// Setup configuration (Initial Parameters comes from web.xml)
		try
		{
			setupConfiguration( evt.getServletContext() );
		}
		catch (Exception e)
		{
			logger.error( "Failed to setup configuration", e );
			throw new RuntimeException( e );
		}
		
		// Setup message ID Generator
		setupMessageIDGeneration( evt.getServletContext() );
	}
	
	
	protected void loadUseCases(ServletContext context)
	{
		logger.info("Setting up service use-cases");
		int totalUC = 0, availableUC = 0;
		try
		{
			UseCaseMap ucMap = new UseCaseMap( context );
			context.setAttribute( SERVICE_USE_CASES, ucMap );
			totalUC = ucMap.getUseCases().size();
			availableUC = ucMap.getAvailableUseCases().size();
		} 
		catch (Exception e)
		{
			logger.fatal("Service use-case setup failed", e);
		}
		logger.info("Service use-cases are ready: "+ availableUC +" available (" + totalUC + " total)");
	}
	
	
	/** Creates and sets up configuration */
	protected void setupConfiguration(ServletContext context) throws IOException
	{
		File configFile = new File( FrameworkSetup.getUserDirectory("xmlvend/test-client"), "configuration.properties");
		Configuration conf = new FileConfiguration( configFile );
		
		// If config file doesn't exist, create it initially and populate the configuration with
		// parameters from the context and save it
		if (!configFile.exists())
		{
			try
			{
				logger.info("Creating default configuration file (first-time use)");
				configFile.createNewFile();
				
				/* All context init params that start with <CONTEXT_PARAM_CONFIG_PREFIX>
				 * automatically go into default configuration.
				 */ 
				Enumeration params = context.getInitParameterNames();
				while (params.hasMoreElements())
				{
					String p = (String)params.nextElement();
					if (p.startsWith(CONTEXT_PARAM_CONFIG_PREFIX))
					{
						String cKey = p.substring(CONTEXT_PARAM_CONFIG_PREFIX.length());
						String cVal = context.getInitParameter( p );
						conf.setProperty( cKey, cVal );
						logger.debug("Default configuration: " + cKey + "=" + cVal);
					}
				}
				
				conf.save();
				logger.info("Configuration file: " + configFile);
			} catch (Exception e)
			{
				logger.fatal("Failed to create configuration file: " + configFile, e);
			}
		}
		
		// Load Configuration from file
		try
		{
			logger.debug("Reading configuration from: " + configFile);
			conf.load();
			
			// FIXME: Also set system properties for SSL/JCE
			
		} catch (IOException e)
		{
			logger.fatal("Failed to read configuration file: " + configFile, e);
		}
		
		// Save in context
		context.setAttribute(CONFIGURATION, conf);
	}
	
	
	protected void setupMessageIDGeneration( ServletContext context )
	{
		context.setAttribute(MESSAGE_ID_GENERATOR, new MessageIDGenerator());
	}
	
	
	public void contextDestroyed(ServletContextEvent evt)
	{
		// Save Configuration ?
	}
	
	
	/** Name of use-case map in context */
	public static final String SERVICE_USE_CASES = "service.useCases";
	/** Name of global configuration in context */
	public static final String CONFIGURATION = "configuration";
	/** Name of message ID egenrator in context */
	public static final String MESSAGE_ID_GENERATOR = "messageIDGenerator";
	/** Prefix of context init params that go into default configuration */
	public static final String CONTEXT_PARAM_CONFIG_PREFIX = "init.";
	
	private static Logger logger = Logger.getLogger(ClientSetupListener.class);
}
