package za.co.eskom.nrs.xmlvend.client.wsdl;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Map.Entry;
import javax.servlet.ServletContext;
import org.apache.log4j.Logger;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Nodes;
import nu.xom.ParsingException;
import nu.xom.ValidityException;
import nu.xom.XPathContext;

/** Represents the use case structure presented by a WSDL. Provides a convenient
 * listing of use cases, as well as the URIs for the request/response JSPX
 * views.
 * 
 * TODO Write unit tests for this class!
 * 
 * @author dawidl
 * @version 0.1 (Dec 19, 2005)
 */
public class UseCaseMap
{
	/** Constructs a use case map using artifacts in the given Servlet context.
	 * Context must have two init params defined:
	 * 
	 * "service.wsdl" - URI of WSDL
	 * "service.operations.views" - URI of directory where views (JSPX) are
	 * 
	 * @throws IOException, ParsingException or ValidityException if WSDL is faulty
	 */
	public UseCaseMap(ServletContext context) throws IOException, ParsingException, ValidityException
	{
		String wsdl = context.getInitParameter("service.wsdl");
		if (wsdl == null)
			throw new IOException("No context parameter to indicate WSDL (service.wsdl)");
		String viewPath = context.getInitParameter("service.operations.views");
		if (viewPath == null)
			throw new IOException("No context parameter to indicate use case views (service.operations.views)");
		
		logger.debug("Parsing contract: " + wsdl);
		InputStream wsdlResource = context.getResourceAsStream( wsdl );
		
		if (wsdlResource == null)
		{
			logger.error("There is no WSDL called: " + wsdl);
			throw new IOException("WSDL reading failed");
		}
		
		Document wsdlDoc = xomBuilder.build( wsdlResource );
		
		// Get all the operations
		logger.debug("Finding all WSDL operations");
		Nodes operations = wsdlDoc.query( "//wsdl:portType/wsdl:operation", xpContext );
		for (int x = 0; x < operations.size(); x++)
		{
			Element operation = (Element)operations.get(x);
			String name = operation.getAttributeValue("name");
			logger.debug("Operation: " + name);
			
			// Get Suporting information
			Nodes doc = operation.query("wsdl:documentation",xpContext);
			String documentation = null;
			if (doc.size() > 0)
				documentation = ((Element)doc.get(0)).getValue();
			logger.debug("Documetation: " + documentation);
			
			// Get view URIs (and test if they exist)
			String inputView = viewPath + "/" + name + "-in.jspx";
			logger.debug("Checking for input view: " + inputView);
			URL inputR = context.getResource(inputView);
			if (inputR == null)
			{
				logger.debug("No input view for " + name + " (disabling)");
				inputView = null;
			}
			String outputView = viewPath + "/" + name + "-out.jspx";
			logger.debug("Checking for output view: " + outputView);
			URL outputR = context.getResource(outputView);
			if (outputR == null)
			{
				logger.debug("No output view for " + name + " (will display raw data)");
				outputView = null;
			}
			
			// Create use case
			UseCase u = new UseCase(name,documentation,inputView,outputView);
			useCases.put(name, u);
			
			// Log
			if (u.isAvailable())
			{
				logger.debug("Available use-case: " + u.getName());
			}
		}
	}
	
	
	/** Get a use case by name (as per WSDL) */
	public UseCase getUseCaseByName( String useCaseName )
	{
		return useCases.get(useCaseName);
	}
	
	
	/** Get all use cases (sorted alphabetically by name) */
	public SortedSet<UseCase> getUseCases()
	{
		// Create a view as a sorted set
		SortedSet<UseCase> us = new TreeSet<UseCase>();
		
		// Add all use cases to set
		for (Entry<String,UseCase> u : useCases.entrySet())
		{
			us.add(u.getValue());
		}
		
		return us;
	}
	
	/** Get only the available use cases (sorted alphabetically by name) */
	public SortedSet<UseCase> getAvailableUseCases()
	{
		// Create a view as a sorted set
		SortedSet<UseCase> us = new TreeSet<UseCase>();
		
		// Add all use cases to set
		for (Entry<String,UseCase> u : useCases.entrySet())
		{
			if (u.getValue().isAvailable())
			{
				us.add(u.getValue());
			}
		}
		
		return us;
	}
	
	
	private Map<String,UseCase> useCases = new HashMap<String,UseCase>();
	public static final String CONTEXT_PARAM_CONTRACT = "service.wsdl";
	public static final String CONTEXT_PARAM_VIEWS = "service.operations.views";
	public static final String WSDL_NAMESPACE = "http://schemas.xmlsoap.org/wsdl/";
	private Builder xomBuilder = new Builder();
	private static final XPathContext xpContext = new XPathContext("wsdl",WSDL_NAMESPACE);
	private static final Logger logger = Logger.getLogger(UseCaseMap.class);
	
	
	/** Represents a single use case */
	public class UseCase implements Comparable<UseCase>
	{
		/** Protected constructor */
		protected UseCase(String name, String documentation, String inputView, String outputView)
		{
			this.name = name;
			// TODO Compute friendly camelcase name
			this.label = name;
			this.documentation = documentation;
			this.inputView = inputView;
			this.outputView = outputView;
		}
		
		/** Retrieve the raw name of the operation */
		public String getName()
		{
			return name;
		}
		
		/** Retrieve the name as to be used on UI components */
		public String getLabel()
		{
			return label;
		}
		
		/** Retrieve help / documentation text */
		public String getDocumentation()
		{
			return documentation;
		}
		
		/** Indicates whether the use case is supported (be there being a valid
		 * UI for necessary operations, e.g. input, output.
		 *  TODO Add necessary logic to tie this up with message exchange pattern
		 */
		public boolean isAvailable()
		{
			return inputView != null;
		}
		
		/** The URI of the request (input) view */
		public String getInputView()
		{
			return inputView;
		}
		
		/** The URI of the response (output) view */
		public String getOutputView()
		{
			return outputView;
		}
		
		/** Comparison based on alphabetic first name (case insensitive) */
		public int compareTo(UseCase u)
		{
			if (u == null || u.getName() == null)
				return 0;
			
			return getName().toLowerCase().compareTo( u.getName().toLowerCase() );
		}
		
		/** Equality based on name only */
		@Override
		public boolean equals(Object o)
		{
			if (this == o)
				return true;
			
			if (o == null || !(o instanceof UseCase))
				return false;
			
			UseCase other = (UseCase)o;
			if (other.getName() == null)
				return false;
			return other.getName().equals( this.getName() );
		}
		
		private String name, label, documentation, inputView, outputView;
		private boolean available;
	}
}